#include "FunctionControls.h"

#ifdef __APPLE__
#include <GLUT/glut.h>
#include <OpenGL/gl.h>
#else
#include <GL/glut.h>
#include <GL/gl.h>
#endif

#include <math.h>
#include <stdlib.h>

#include "Matrix.h"
#include "NumberField.h"
#include "Vector.h"
#include "Utilities.h"

enum {
	TRANSLATE_X = 0,
	TRANSLATE_Y,
	TRANSLATE_Z,
	SCALE_X,
	SCALE_Y,
	SCALE_Z,
	ROTATE_X,
	ROTATE_Y,
	ROTATE_Z,
	ROTATE_ANGLE,
	SHEARX_Y,
	SHEARX_Z,
	SHEARY_X,
	SHEARY_Z,
	SHEARZ_X,
	SHEARZ_Y,
	PERSPECTIVE_FOVY,
	PERSPECTIVE_ASPECT,
	PERSPECTIVE_ZNEAR,
	PERSPECTIVE_ZFAR,
	NUMBER_OF_FIELDS
};

static NumberField numberFields[NUMBER_OF_FIELDS];
static int editNumberField = -1;

static void callback(void * context, float newValue) {
	glutPostRedisplay();
}

static void vectorCallback(void * context, float newValue) {
	Vector vector;
	
	vector.x = numberFields[ROTATE_X].value;
	vector.y = numberFields[ROTATE_Y].value;
	vector.z = numberFields[ROTATE_Z].value;
	if (Vector_magnitudeSquared(vector) > 0.0001f) {
		Vector_normalize(&vector);
		numberFields[ROTATE_X].value = vector.x;
		numberFields[ROTATE_Y].value = vector.y;
		numberFields[ROTATE_Z].value = vector.z;
	}
	glutPostRedisplay();
}

void FunctionControls_init() {
	numberFields[TRANSLATE_X] = NumberField_create(true, 0.0f, 0.0025f, callback, NULL, 0, 198, 149);
	numberFields[TRANSLATE_Y] = NumberField_create(true, 0.0f, 0.0025f, callback, NULL, 0, 274, 149);
	numberFields[TRANSLATE_Z] = NumberField_create(true, 0.0f, 0.0025f, callback, NULL, 0, 350, 149);
	
	numberFields[SCALE_X] = NumberField_create(true, 1.0f, 0.0025f, callback, NULL, 0, 166, 131);
	numberFields[SCALE_Y] = NumberField_create(true, 1.0f, 0.0025f, callback, NULL, 0, 242, 131);
	numberFields[SCALE_Z] = NumberField_create(true, 1.0f, 0.0025f, callback, NULL, 0, 318, 131);
	
	numberFields[ROTATE_X] = NumberField_create(true, 0.0f, 0.0025f, vectorCallback, NULL, 0, 318, 113);
	numberFields[ROTATE_Y] = NumberField_create(true, 0.0f, 0.0025f, vectorCallback, NULL, 0, 394, 113);
	numberFields[ROTATE_Z] = NumberField_create(true, 0.0f, 0.0025f, vectorCallback, NULL, 0, 470, 113);
	numberFields[ROTATE_ANGLE] = NumberField_create(true, 0.0f, 0.0025f, callback, NULL, 0, 568, 113);
	
	numberFields[SHEARX_Y] = NumberField_create(true, 0.0f, 0.0025f, callback, NULL, 0, 174, 95);
	numberFields[SHEARX_Z] = NumberField_create(true, 0.0f, 0.0025f, callback, NULL, 0, 250, 95);
	
	numberFields[SHEARY_X] = NumberField_create(true, 0.0f, 0.0025f, callback, NULL, 0, 174, 77);
	numberFields[SHEARY_Z] = NumberField_create(true, 0.0f, 0.0025f, callback, NULL, 0, 250, 77);
	
	numberFields[SHEARZ_X] = NumberField_create(true, 0.0f, 0.0025f, callback, NULL, 0, 174, 59);
	numberFields[SHEARZ_Y] = NumberField_create(true, 0.0f, 0.0025f, callback, NULL, 0, 250, 59);
	
	numberFields[PERSPECTIVE_FOVY]   = NumberField_create(true, 0.0f, 0.0025f, callback, NULL, 0, 214, 41);
	numberFields[PERSPECTIVE_ASPECT] = NumberField_create(true, 0.0f, 0.0025f, callback, NULL, 0, 289, 41);
	numberFields[PERSPECTIVE_ZNEAR]  = NumberField_create(true, 0.0f, 0.0025f, callback, NULL, 0, 364, 41);
	numberFields[PERSPECTIVE_ZFAR]   = NumberField_create(true, 0.0f, 0.0025f, callback, NULL, 0, 439, 41);
}

bool FunctionControls_mouseDown(int x, int y) {
	int numberFieldIndex;
	
	for (numberFieldIndex = 0; numberFieldIndex < NUMBER_OF_FIELDS; numberFieldIndex++) {
		if (NumberField_hitTest(&numberFields[numberFieldIndex], x, y)) {
			editNumberField = numberFieldIndex;
			return true;
		}
	}
	return false;
}

void FunctionControls_mouseUp(int x, int y) {
	editNumberField = -1;
}

void FunctionControls_mouseDragged(int x, int y) {
	if (editNumberField != -1) {
		NumberField_mouseDragged(&numberFields[editNumberField], x, y);
	}
}

void FunctionControls_draw() {
	int numberFieldIndex;
	
	for (numberFieldIndex = 0; numberFieldIndex < NUMBER_OF_FIELDS; numberFieldIndex++) {
		NumberField_draw(&numberFields[numberFieldIndex], numberFieldIndex == editNumberField);
	}
	
	drawGlutString(56, 152, "Matrix_translate(");
	drawGlutString(424, 152, ");");
	
	drawGlutString(56, 134, "Matrix_scale(");
	drawGlutString(392, 134, ");");
	
	drawGlutString(56, 116, "Matrix_rotate(Vector_withValues(");
	drawGlutString(546, 116, "),");
	drawGlutString(644, 116, ");");
	
	drawGlutString(56, 98, "Matrix_shearX(");
	drawGlutString(324, 98, ");");
	
	drawGlutString(56, 80, "Matrix_shearY(");
	drawGlutString(324, 80, ");");
	
	drawGlutString(56, 62, "Matrix_shearZ(");
	drawGlutString(324, 62, ");");
	
	drawGlutString(56, 44, "Matrix_perspective(");
	drawGlutString(512, 44, ");");
	
	drawGlutString(56, 26, "Matrix_invert();");
	drawGlutString(56, 8, "Matrix_transpose();");
}

void FunctionControls_setEditFieldValue(float value) {
	if (editNumberField != -1) {
		NumberField_setValue(&numberFields[editNumberField], value);
	}
}

float FunctionControls_getEditFieldValue() {
	if (editNumberField != -1) {
		return numberFields[editNumberField].value;
	}
	return 0.0f;
}

struct Matrix FunctionControls_getTranslationMatrix() {
	return Matrix_translated(Matrix_identity(), numberFields[TRANSLATE_X].value, numberFields[TRANSLATE_Y].value, numberFields[TRANSLATE_Z].value);
}

struct Matrix FunctionControls_getScalingMatrix() {
	return Matrix_scaled(Matrix_identity(), numberFields[SCALE_X].value, numberFields[SCALE_Y].value, numberFields[SCALE_Z].value);
}

struct Matrix FunctionControls_getRotationMatrix() {
	return Matrix_rotated(Matrix_identity(), Vector_withValues(numberFields[ROTATE_X].value, numberFields[ROTATE_Y].value, numberFields[ROTATE_Z].value), numberFields[ROTATE_ANGLE].value);
}

struct Matrix FunctionControls_getShearXMatrix() {
	return Matrix_shearedX(Matrix_identity(), numberFields[SHEARX_Y].value, numberFields[SHEARX_Z].value);
}

struct Matrix FunctionControls_getShearYMatrix() {
	return Matrix_shearedY(Matrix_identity(), numberFields[SHEARY_X].value, numberFields[SHEARY_Z].value);
}

struct Matrix FunctionControls_getShearZMatrix() {
	return Matrix_shearedZ(Matrix_identity(), numberFields[SHEARZ_X].value, numberFields[SHEARZ_Y].value);
}

struct Matrix FunctionControls_getPerspectiveMatrix() {
	return Matrix_perspective(Matrix_identity(), numberFields[PERSPECTIVE_FOVY].value, numberFields[PERSPECTIVE_ASPECT].value, numberFields[PERSPECTIVE_ZNEAR].value, numberFields[PERSPECTIVE_ZFAR].value);
}
