#include "MatrixControls.h"

#ifdef __APPLE__
#include <OpenGL/gl.h>
#include <GLUT/glut.h>
#else
#include <GL/gl.h>
#include <GL/glut.h>
#endif

#include <stdlib.h>

#include "Matrix.h"
#include "NumberField.h"
#include "Utilities.h"

enum {
	FIELD_MAIN_0,
	FIELD_MAIN_1,
	FIELD_MAIN_2,
	FIELD_MAIN_3,
	FIELD_MAIN_4,
	FIELD_MAIN_5,
	FIELD_MAIN_6,
	FIELD_MAIN_7,
	FIELD_MAIN_8,
	FIELD_MAIN_9,
	FIELD_MAIN_10,
	FIELD_MAIN_11,
	FIELD_MAIN_12,
	FIELD_MAIN_13,
	FIELD_MAIN_14,
	FIELD_MAIN_15,
	FIELD_OPERAND_0,
	FIELD_OPERAND_1,
	FIELD_OPERAND_2,
	FIELD_OPERAND_3,
	FIELD_OPERAND_4,
	FIELD_OPERAND_5,
	FIELD_OPERAND_6,
	FIELD_OPERAND_7,
	FIELD_OPERAND_8,
	FIELD_OPERAND_9,
	FIELD_OPERAND_10,
	FIELD_OPERAND_11,
	FIELD_OPERAND_12,
	FIELD_OPERAND_13,
	FIELD_OPERAND_14,
	FIELD_OPERAND_15,
	NUMBER_OF_FIELDS
};

static void (* matrixChangeCallback)(struct Matrix);
static NumberField numberFields[NUMBER_OF_FIELDS];
static int dragNumberField = -1;

static void numberFieldChangeCallback(void * context, float newValue) {
	(*matrixChangeCallback)(Matrix_withValues(numberFields[FIELD_MAIN_0].value, numberFields[FIELD_MAIN_4].value, numberFields[FIELD_MAIN_8].value,  numberFields[FIELD_MAIN_12].value,
	                                          numberFields[FIELD_MAIN_1].value, numberFields[FIELD_MAIN_5].value, numberFields[FIELD_MAIN_9].value,  numberFields[FIELD_MAIN_13].value,
	                                          numberFields[FIELD_MAIN_2].value, numberFields[FIELD_MAIN_6].value, numberFields[FIELD_MAIN_10].value, numberFields[FIELD_MAIN_14].value,
	                                          numberFields[FIELD_MAIN_3].value, numberFields[FIELD_MAIN_7].value, numberFields[FIELD_MAIN_11].value, numberFields[FIELD_MAIN_15].value));
}

static void postRedisplayCallback(void * context, float newValue) {
	glutPostRedisplay();
}

void MatrixControls_init(void (* inMatrixChangeCallback)(struct Matrix)) {
	matrixChangeCallback = inMatrixChangeCallback;
	
	numberFields[FIELD_MAIN_0]  = NumberField_create(true, 0.0f, 0.0025f, numberFieldChangeCallback, NULL, 1, 5, 79);
	numberFields[FIELD_MAIN_4]  = NumberField_create(true, 0.0f, 0.0025f, numberFieldChangeCallback, NULL, 2, 80, 79);
	numberFields[FIELD_MAIN_8]  = NumberField_create(true, 0.0f, 0.0025f, numberFieldChangeCallback, NULL, 3, 155, 79);
	numberFields[FIELD_MAIN_12] = NumberField_create(true, 0.0f, 0.0025f, numberFieldChangeCallback, NULL, 4, 230, 79);
	numberFields[FIELD_MAIN_1]  = NumberField_create(true, 0.0f, 0.0025f, numberFieldChangeCallback, NULL, 1, 5, 61);
	numberFields[FIELD_MAIN_5]  = NumberField_create(true, 0.0f, 0.0025f, numberFieldChangeCallback, NULL, 2, 80, 61);
	numberFields[FIELD_MAIN_9]  = NumberField_create(true, 0.0f, 0.0025f, numberFieldChangeCallback, NULL, 3, 155, 61);
	numberFields[FIELD_MAIN_13] = NumberField_create(true, 0.0f, 0.0025f, numberFieldChangeCallback, NULL, 4, 230, 61);
	numberFields[FIELD_MAIN_2]  = NumberField_create(true, 0.0f, 0.0025f, numberFieldChangeCallback, NULL, 1, 5, 43);
	numberFields[FIELD_MAIN_6]  = NumberField_create(true, 0.0f, 0.0025f, numberFieldChangeCallback, NULL, 2, 80, 43);
	numberFields[FIELD_MAIN_10] = NumberField_create(true, 0.0f, 0.0025f, numberFieldChangeCallback, NULL, 3, 155, 43);
	numberFields[FIELD_MAIN_14] = NumberField_create(true, 0.0f, 0.0025f, numberFieldChangeCallback, NULL, 4, 230, 43);
	numberFields[FIELD_MAIN_3]  = NumberField_create(true, 0.0f, 0.0025f, numberFieldChangeCallback, NULL, 5, 5, 25);
	numberFields[FIELD_MAIN_7]  = NumberField_create(true, 0.0f, 0.0025f, numberFieldChangeCallback, NULL, 5, 80, 25);
	numberFields[FIELD_MAIN_11] = NumberField_create(true, 0.0f, 0.0025f, numberFieldChangeCallback, NULL, 5, 155, 25);
	numberFields[FIELD_MAIN_15] = NumberField_create(true, 0.0f, 0.0025f, numberFieldChangeCallback, NULL, 6, 230, 25);
	
	numberFields[FIELD_OPERAND_0]  = NumberField_create(true, 1.0f, 0.0025f, postRedisplayCallback, NULL, 1, 502, 79);
	numberFields[FIELD_OPERAND_4]  = NumberField_create(true, 0.0f, 0.0025f, postRedisplayCallback, NULL, 2, 577, 79);
	numberFields[FIELD_OPERAND_8]  = NumberField_create(true, 0.0f, 0.0025f, postRedisplayCallback, NULL, 3, 652, 79);
	numberFields[FIELD_OPERAND_12] = NumberField_create(true, 0.0f, 0.0025f, postRedisplayCallback, NULL, 4, 727, 79);
	numberFields[FIELD_OPERAND_1]  = NumberField_create(true, 0.0f, 0.0025f, postRedisplayCallback, NULL, 1, 502, 61);
	numberFields[FIELD_OPERAND_5]  = NumberField_create(true, 1.0f, 0.0025f, postRedisplayCallback, NULL, 2, 577, 61);
	numberFields[FIELD_OPERAND_9]  = NumberField_create(true, 0.0f, 0.0025f, postRedisplayCallback, NULL, 3, 652, 61);
	numberFields[FIELD_OPERAND_13] = NumberField_create(true, 0.0f, 0.0025f, postRedisplayCallback, NULL, 4, 727, 61);
	numberFields[FIELD_OPERAND_2]  = NumberField_create(true, 0.0f, 0.0025f, postRedisplayCallback, NULL, 1, 502, 43);
	numberFields[FIELD_OPERAND_6]  = NumberField_create(true, 0.0f, 0.0025f, postRedisplayCallback, NULL, 2, 577, 43);
	numberFields[FIELD_OPERAND_10] = NumberField_create(true, 1.0f, 0.0025f, postRedisplayCallback, NULL, 3, 652, 43);
	numberFields[FIELD_OPERAND_14] = NumberField_create(true, 0.0f, 0.0025f, postRedisplayCallback, NULL, 4, 727, 43);
	numberFields[FIELD_OPERAND_3]  = NumberField_create(true, 0.0f, 0.0025f, postRedisplayCallback, NULL, 5, 502, 25);
	numberFields[FIELD_OPERAND_7]  = NumberField_create(true, 0.0f, 0.0025f, postRedisplayCallback, NULL, 5, 577, 25);
	numberFields[FIELD_OPERAND_11] = NumberField_create(true, 0.0f, 0.0025f, postRedisplayCallback, NULL, 5, 652, 25);
	numberFields[FIELD_OPERAND_15] = NumberField_create(true, 1.0f, 0.0025f, postRedisplayCallback, NULL, 6, 727, 25);
}

void MatrixControls_setMainMatrix(struct Matrix inMatrix) {
	numberFields[FIELD_MAIN_0].value  = inMatrix.m[0];
	numberFields[FIELD_MAIN_1].value  = inMatrix.m[1];
	numberFields[FIELD_MAIN_2].value  = inMatrix.m[2];
	numberFields[FIELD_MAIN_3].value  = inMatrix.m[3];
	numberFields[FIELD_MAIN_4].value  = inMatrix.m[4];
	numberFields[FIELD_MAIN_5].value  = inMatrix.m[5];
	numberFields[FIELD_MAIN_6].value  = inMatrix.m[6];
	numberFields[FIELD_MAIN_7].value  = inMatrix.m[7];
	numberFields[FIELD_MAIN_8].value  = inMatrix.m[8];
	numberFields[FIELD_MAIN_9].value  = inMatrix.m[9];
	numberFields[FIELD_MAIN_10].value = inMatrix.m[10];
	numberFields[FIELD_MAIN_11].value = inMatrix.m[11];
	numberFields[FIELD_MAIN_12].value = inMatrix.m[12];
	numberFields[FIELD_MAIN_13].value = inMatrix.m[13];
	numberFields[FIELD_MAIN_14].value = inMatrix.m[14];
	numberFields[FIELD_MAIN_15].value = inMatrix.m[15];
}

void MatrixControls_setOperandMatrix(struct Matrix inMatrix) {
	numberFields[FIELD_OPERAND_0].value  = inMatrix.m[0];
	numberFields[FIELD_OPERAND_1].value  = inMatrix.m[1];
	numberFields[FIELD_OPERAND_2].value  = inMatrix.m[2];
	numberFields[FIELD_OPERAND_3].value  = inMatrix.m[3];
	numberFields[FIELD_OPERAND_4].value  = inMatrix.m[4];
	numberFields[FIELD_OPERAND_5].value  = inMatrix.m[5];
	numberFields[FIELD_OPERAND_6].value  = inMatrix.m[6];
	numberFields[FIELD_OPERAND_7].value  = inMatrix.m[7];
	numberFields[FIELD_OPERAND_8].value  = inMatrix.m[8];
	numberFields[FIELD_OPERAND_9].value  = inMatrix.m[9];
	numberFields[FIELD_OPERAND_10].value = inMatrix.m[10];
	numberFields[FIELD_OPERAND_11].value = inMatrix.m[11];
	numberFields[FIELD_OPERAND_12].value = inMatrix.m[12];
	numberFields[FIELD_OPERAND_13].value = inMatrix.m[13];
	numberFields[FIELD_OPERAND_14].value = inMatrix.m[14];
	numberFields[FIELD_OPERAND_15].value = inMatrix.m[15];
}

Matrix MatrixControls_getOperandMatrix() {
	return Matrix_withValues(numberFields[FIELD_OPERAND_0].value, numberFields[FIELD_OPERAND_4].value, numberFields[FIELD_OPERAND_8].value,  numberFields[FIELD_OPERAND_12].value,
	                         numberFields[FIELD_OPERAND_1].value, numberFields[FIELD_OPERAND_5].value, numberFields[FIELD_OPERAND_9].value,  numberFields[FIELD_OPERAND_13].value,
	                         numberFields[FIELD_OPERAND_2].value, numberFields[FIELD_OPERAND_6].value, numberFields[FIELD_OPERAND_10].value, numberFields[FIELD_OPERAND_14].value,
	                         numberFields[FIELD_OPERAND_3].value, numberFields[FIELD_OPERAND_7].value, numberFields[FIELD_OPERAND_11].value, numberFields[FIELD_OPERAND_15].value);
}

void MatrixControls_setEditFieldValue(float value) {
	if (dragNumberField != -1) {
		NumberField_setValue(&numberFields[dragNumberField], value);
	}
}

float MatrixControls_getEditFieldValue() {
	if (dragNumberField != -1) {
		return numberFields[dragNumberField].value;
	}
	return 0.0f;
}

bool MatrixControls_mouseDown(int x, int y) {
	int numberFieldIndex;
	
	for (numberFieldIndex = 0; numberFieldIndex < NUMBER_OF_FIELDS; numberFieldIndex++) {
		if (NumberField_hitTest(&numberFields[numberFieldIndex], x, y)) {
			dragNumberField = numberFieldIndex;
			return true;
		}
	}
	return false;
}

void MatrixControls_mouseUp(int x, int y) {
	dragNumberField = -1;
}

void MatrixControls_mouseDragged(int x, int y) {
	if (dragNumberField != -1) {
		NumberField_mouseDragged(&numberFields[dragNumberField], x, y);
	}
}

void MatrixControls_draw(int autoMultiplyMode) {
	int numberFieldIndex;
	
	for (numberFieldIndex = 0; numberFieldIndex < NUMBER_OF_FIELDS; numberFieldIndex++) {
		NumberField_draw(&numberFields[numberFieldIndex], numberFieldIndex == dragNumberField);
	}
	
	drawGlutString(70, 100, "left (current matrix)");
	switch (autoMultiplyMode) {
		case 0:
			drawGlutString(587, 100, "right (not visible)");
			break;
		case 1:
			drawGlutString(531, 100, "right (multiplies after left)");
			break;
		case 2:
			drawGlutString(527, 100, "right (multiplies before left)");
			break;
	}
}
