#ifdef __APPLE__
#include <GLUT/glut.h>
#include <OpenGL/gl.h>
#include <OpenGL/OpenGL.h>
#else
#include <GL/glut.h>
#include <GL/gl.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <math.h>

#include "Button.h"
#include "DisplayObjects.h"
#include "FunctionControls.h"
#include "Matrix.h"
#include "MatrixControls.h"
#include "Quaternion.h"
#include "Utilities.h"
#include "Vector.h"

enum {
	RESET_TO_IDENTITY_BUTTON_INDEX,
	MULTIPLY_MODE_BUTTON_INDEX,
	MULTIPLY_LEFT_BY_RIGHT_BUTTON_INDEX,
	MULTIPLY_RIGHT_BY_LEFT_BUTTON_INDEX,
	SWAP_BUTTON_INDEX,
	RESET_OPERAND_TO_IDENTITY_BUTTON_INDEX,
	ROTATE_MULTIPLY_ORDER_BUTTON_INDEX,
	SHOW_FUNCTIONS_BUTTON_INDEX,
	APPLY_TRANSLATE_BUTTON_INDEX,
	APPLY_SCALE_BUTTON_INDEX,
	APPLY_ROTATE_BUTTON_INDEX,
	APPLY_SHEARX_BUTTON_INDEX,
	APPLY_SHEARY_BUTTON_INDEX,
	APPLY_SHEARZ_BUTTON_INDEX,
	APPLY_PERSPECTIVE_BUTTON_INDEX,
	APPLY_INVERT_BUTTON_INDEX,
	APPLY_TRANSPOSE_BUTTON_INDEX,
	NUMBER_OF_BUTTONS
};

static Matrix matrix;

static bool mouseIsDown = false;
static int lastMouseX = 0;
static int lastMouseY = 0;
static bool draggingNumberField = false;
static bool draggingFunctionField = false;

static Button buttons[NUMBER_OF_BUTTONS];
static Button * selectedButton;

static bool showingFunctions = false;
static bool multiplyRotationFirst = false;
static int autoMultiplyMode = 0;

static void initGL() {
	GLint viewport[4];
	float matSpecular[] = {1.0f, 1.0f, 1.0f, 1.0f};
	float ambient[] = {0.1f, 0.1f, 0.1f, 1.0f};
	
#ifdef __APPLE__
	GLint VBL = 1;
	CGLSetParameter(CGLGetCurrentContext(), kCGLCPSwapInterval, &VBL);
#endif
	
	glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
	glEnable(GL_DEPTH_TEST);
	
	glEnable(GL_LIGHTING);
	glEnable(GL_NORMALIZE);
	glEnable(GL_COLOR_MATERIAL);
	glEnable(GL_LIGHT0);
	glEnable(GL_LIGHT1);
	
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, ambient);
	
	glMaterialf(GL_FRONT, GL_SHININESS, 32.0);
	glMaterialfv(GL_FRONT, GL_SPECULAR, matSpecular);
	
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glGetIntegerv(GL_VIEWPORT, viewport);
	gluPerspective(60, (float) viewport[2] / (float) viewport[3], 1.5, 400.0);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
}

static void matrixChangedCallback(Matrix value) {
	matrix = value;
	MatrixControls_setMainMatrix(matrix);
	glutPostRedisplay();
}

static void initPlayground() {
	matrix = Matrix_identity();
	MatrixControls_init(matrixChangedCallback);
	MatrixControls_setMainMatrix(matrix);
	FunctionControls_init();
	buttons[RESET_TO_IDENTITY_BUTTON_INDEX] = Button_create("Reset to identity ^", 5, 5, 158);
	buttons[ROTATE_MULTIPLY_ORDER_BUTTON_INDEX] = Button_create("Drag mode: current * rotation", 200, 5, 238);
	buttons[MULTIPLY_MODE_BUTTON_INDEX] = Button_create(" Show:  left only", 326, 25, 150);
	buttons[SWAP_BUTTON_INDEX] = Button_create("<- Swap ->", 357, 43, 86);
	buttons[MULTIPLY_LEFT_BY_RIGHT_BUTTON_INDEX] = Button_create("Multiply left by right", 310, 79, 182);
	buttons[MULTIPLY_RIGHT_BY_LEFT_BUTTON_INDEX] = Button_create("Multiply right by left", 310, 61, 182);
	buttons[RESET_OPERAND_TO_IDENTITY_BUTTON_INDEX] = Button_create("Reset to identity ^", 502, 5, 158);
	buttons[SHOW_FUNCTIONS_BUTTON_INDEX] = Button_create("Functions ->", 700, 5, 102);
	buttons[APPLY_TRANSLATE_BUTTON_INDEX] = Button_create("Apply", 5, 149, 46);
	buttons[APPLY_SCALE_BUTTON_INDEX] = Button_create("Apply", 5, 131, 46);
	buttons[APPLY_ROTATE_BUTTON_INDEX] = Button_create("Apply", 5, 113, 46);
	buttons[APPLY_SHEARX_BUTTON_INDEX] = Button_create("Apply", 5, 95, 46);
	buttons[APPLY_SHEARY_BUTTON_INDEX] = Button_create("Apply", 5, 77, 46);
	buttons[APPLY_SHEARZ_BUTTON_INDEX] = Button_create("Apply", 5, 59, 46);
	buttons[APPLY_PERSPECTIVE_BUTTON_INDEX] = Button_create("Apply", 5, 41, 46);
	buttons[APPLY_INVERT_BUTTON_INDEX] = Button_create("Apply", 5, 23, 46);
	buttons[APPLY_TRANSPOSE_BUTTON_INDEX] = Button_create("Apply", 5, 5, 46);
}

static void displayFunc() {
	float lightPosition0[] = {-1.0f, 2.0f, 0.0f, 0.0f}, lightPosition1[] = {2.0f, -5.0f, 2.0f, 0.0f};
	float lightDiffuse[] = {0.85f, 0.85f, 0.85f, 1.0f};
	GLint viewport[4];
	int currentTime;
	
	currentTime = glutGet(GLUT_ELAPSED_TIME);
	glGetIntegerv(GL_VIEWPORT, viewport);
	
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	
	glLoadIdentity();
	glTranslatef(0.0f, 0.0f, -5.5f);
	
	glLightfv(GL_LIGHT0, GL_POSITION, lightPosition0);
	glLightfv(GL_LIGHT1, GL_POSITION, lightPosition1);
	glLightfv(GL_LIGHT1, GL_DIFFUSE, lightDiffuse);
	
	glViewport(0, 0, viewport[2] / 3, viewport[3]);
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glTranslatef(0.0f, 0.15f, 0.0f);
	gluPerspective(60, (float) viewport[2] / 3 / (float) viewport[3], 1.5, 400.0);
	glMatrixMode(GL_MODELVIEW);
	drawWireMesh();
	glPushMatrix();
	switch (autoMultiplyMode) {
		case 0:
			applyMatrix(matrix);
			break;
		case 1:
			applyMatrix(matrix);
			applyMatrix(MatrixControls_getOperandMatrix());
			break;
		case 2:
			applyMatrix(MatrixControls_getOperandMatrix());
			applyMatrix(matrix);
			break;
	}
	drawAxisColoredSphere();
	glPopMatrix();
	
	glViewport(viewport[2] / 3, 0, viewport[2] / 3, viewport[3]);
	drawWireMesh();
	glPushMatrix();
	switch (autoMultiplyMode) {
		case 0:
			applyMatrix(matrix);
			break;
		case 1:
			applyMatrix(matrix);
			applyMatrix(MatrixControls_getOperandMatrix());
			break;
		case 2:
			applyMatrix(MatrixControls_getOperandMatrix());
			applyMatrix(matrix);
			break;
	}
	glutSolidTeapot(1.0f);
	glPopMatrix();
	
	glViewport(viewport[2] / 3 * 2, 0, viewport[2] / 3, viewport[3]);
	drawWireMesh();
	glPushMatrix();
	switch (autoMultiplyMode) {
		case 0:
			applyMatrix(matrix);
			break;
		case 1:
			applyMatrix(matrix);
			applyMatrix(MatrixControls_getOperandMatrix());
			break;
		case 2:
			applyMatrix(MatrixControls_getOperandMatrix());
			applyMatrix(matrix);
			break;
	}
	drawAxisCross();
	glPopMatrix();
	
	glViewport(0, 0, viewport[2], viewport[3]);
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	
	glMatrixMode(GL_PROJECTION);
	glPushMatrix();
	glLoadIdentity();
	glOrtho(0, viewport[2], 0, viewport[3], -1, 1);
	glMatrixMode(GL_MODELVIEW);
	glPushMatrix();
	glLoadIdentity();
	glDisable(GL_LIGHTING);
	glDisable(GL_DEPTH_TEST);
	
	if (showingFunctions) {
		FunctionControls_draw();
		Button_draw(&buttons[SHOW_FUNCTIONS_BUTTON_INDEX], selectedButton == &buttons[SHOW_FUNCTIONS_BUTTON_INDEX]);
		Button_draw(&buttons[APPLY_TRANSLATE_BUTTON_INDEX], selectedButton == &buttons[APPLY_TRANSLATE_BUTTON_INDEX]);
		Button_draw(&buttons[APPLY_SCALE_BUTTON_INDEX], selectedButton == &buttons[APPLY_SCALE_BUTTON_INDEX]);
		Button_draw(&buttons[APPLY_ROTATE_BUTTON_INDEX], selectedButton == &buttons[APPLY_ROTATE_BUTTON_INDEX]);
		Button_draw(&buttons[APPLY_SHEARX_BUTTON_INDEX], selectedButton == &buttons[APPLY_SHEARX_BUTTON_INDEX]);
		Button_draw(&buttons[APPLY_SHEARY_BUTTON_INDEX], selectedButton == &buttons[APPLY_SHEARY_BUTTON_INDEX]);
		Button_draw(&buttons[APPLY_SHEARZ_BUTTON_INDEX], selectedButton == &buttons[APPLY_SHEARZ_BUTTON_INDEX]);
		Button_draw(&buttons[APPLY_PERSPECTIVE_BUTTON_INDEX], selectedButton == &buttons[APPLY_PERSPECTIVE_BUTTON_INDEX]);
		Button_draw(&buttons[APPLY_INVERT_BUTTON_INDEX], selectedButton == &buttons[APPLY_INVERT_BUTTON_INDEX]);
		Button_draw(&buttons[APPLY_TRANSPOSE_BUTTON_INDEX], selectedButton == &buttons[APPLY_TRANSPOSE_BUTTON_INDEX]);
	} else {
		MatrixControls_draw(autoMultiplyMode);
		Button_draw(&buttons[RESET_TO_IDENTITY_BUTTON_INDEX], selectedButton == &buttons[RESET_TO_IDENTITY_BUTTON_INDEX]);
		Button_draw(&buttons[MULTIPLY_MODE_BUTTON_INDEX], selectedButton == &buttons[MULTIPLY_MODE_BUTTON_INDEX]);
		Button_draw(&buttons[MULTIPLY_LEFT_BY_RIGHT_BUTTON_INDEX], selectedButton == &buttons[MULTIPLY_LEFT_BY_RIGHT_BUTTON_INDEX]);
		Button_draw(&buttons[MULTIPLY_RIGHT_BY_LEFT_BUTTON_INDEX], selectedButton == &buttons[MULTIPLY_RIGHT_BY_LEFT_BUTTON_INDEX]);
		Button_draw(&buttons[SWAP_BUTTON_INDEX], selectedButton == &buttons[SWAP_BUTTON_INDEX]);
		Button_draw(&buttons[RESET_OPERAND_TO_IDENTITY_BUTTON_INDEX], selectedButton == &buttons[RESET_OPERAND_TO_IDENTITY_BUTTON_INDEX]);
		Button_draw(&buttons[SHOW_FUNCTIONS_BUTTON_INDEX], selectedButton == &buttons[SHOW_FUNCTIONS_BUTTON_INDEX]);
		Button_draw(&buttons[ROTATE_MULTIPLY_ORDER_BUTTON_INDEX], selectedButton == &buttons[ROTATE_MULTIPLY_ORDER_BUTTON_INDEX]);
		drawGlutString(5, 145, "These values can be modified by dragging up or down with your mouse.");
		drawGlutString(5, 132, "For greater precision, start the drag closer to the left side of the field.");
		drawGlutString(5, 119, "While dragging, type a number (0-9) to set the value of the field. You can also negate it with -.");
	}
	
	drawGlutString(viewport[2] / 2 - 220, viewport[3] - 13, "Each of the objects below is transformed by the matrix.");
	drawGlutString(viewport[2] / 2 - 284, viewport[3] - 26, "You can drag your mouse in the scene to apply a rotation to the matrix.");
	
	glMatrixMode(GL_PROJECTION);
	glPopMatrix();
	glMatrixMode(GL_MODELVIEW);
	glPopMatrix();
	glEnable(GL_LIGHTING);
	glEnable(GL_DEPTH_TEST);
	
	glutSwapBuffers();
}

void reshapeFunc(int width, int height) {
	glViewport(0, 0, width, height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(60, (float) width / (float) height, 1.5, 400.0);
	glMatrixMode(GL_MODELVIEW);
#ifndef __APPLE__
	displayFunc();
#endif
}

void mouseFunc(int button, int state, int x, int y) {
	if (button == GLUT_LEFT_BUTTON) {
		GLint viewport[4];
		
		glGetIntegerv(GL_VIEWPORT, viewport);
		y = viewport[3] - y;
		
		mouseIsDown = state == GLUT_DOWN;
		lastMouseX = x;
		lastMouseY = y;
		
		if (mouseIsDown) {
			if (showingFunctions) {
				if (FunctionControls_mouseDown(x, y)) {
					draggingFunctionField = true;
					
				} else if (Button_hitTest(&buttons[SHOW_FUNCTIONS_BUTTON_INDEX], x, y)) {
					selectedButton = &buttons[SHOW_FUNCTIONS_BUTTON_INDEX];
					
				} else if (Button_hitTest(&buttons[APPLY_TRANSLATE_BUTTON_INDEX], x, y)) {
					selectedButton = &buttons[APPLY_TRANSLATE_BUTTON_INDEX];
					
				} else if (Button_hitTest(&buttons[APPLY_SCALE_BUTTON_INDEX], x, y)) {
					selectedButton = &buttons[APPLY_SCALE_BUTTON_INDEX];
					
				} else if (Button_hitTest(&buttons[APPLY_ROTATE_BUTTON_INDEX], x, y)) {
					selectedButton = &buttons[APPLY_ROTATE_BUTTON_INDEX];
					
				} else if (Button_hitTest(&buttons[APPLY_SHEARX_BUTTON_INDEX], x, y)) {
					selectedButton = &buttons[APPLY_SHEARX_BUTTON_INDEX];
					
				} else if (Button_hitTest(&buttons[APPLY_SHEARY_BUTTON_INDEX], x, y)) {
					selectedButton = &buttons[APPLY_SHEARY_BUTTON_INDEX];
					
				} else if (Button_hitTest(&buttons[APPLY_SHEARZ_BUTTON_INDEX], x, y)) {
					selectedButton = &buttons[APPLY_SHEARZ_BUTTON_INDEX];
					
				} else if (Button_hitTest(&buttons[APPLY_PERSPECTIVE_BUTTON_INDEX], x, y)) {
					selectedButton = &buttons[APPLY_PERSPECTIVE_BUTTON_INDEX];
					
				} else if (Button_hitTest(&buttons[APPLY_INVERT_BUTTON_INDEX], x, y)) {
					selectedButton = &buttons[APPLY_INVERT_BUTTON_INDEX];
					
				} else if (Button_hitTest(&buttons[APPLY_TRANSPOSE_BUTTON_INDEX], x, y)) {
					selectedButton = &buttons[APPLY_TRANSPOSE_BUTTON_INDEX];
				}
			} else {
				if (MatrixControls_mouseDown(x, y)) {
					draggingNumberField = true;
					
				} else if (Button_hitTest(&buttons[RESET_TO_IDENTITY_BUTTON_INDEX], x, y)) {
					selectedButton = &buttons[RESET_TO_IDENTITY_BUTTON_INDEX];
					
				} else if (Button_hitTest(&buttons[MULTIPLY_MODE_BUTTON_INDEX], x, y)) {
					selectedButton = &buttons[MULTIPLY_MODE_BUTTON_INDEX];
					
				} else if (Button_hitTest(&buttons[MULTIPLY_LEFT_BY_RIGHT_BUTTON_INDEX], x, y)) {
					selectedButton = &buttons[MULTIPLY_LEFT_BY_RIGHT_BUTTON_INDEX];
					
				} else if (Button_hitTest(&buttons[MULTIPLY_RIGHT_BY_LEFT_BUTTON_INDEX], x, y)) {
					selectedButton = &buttons[MULTIPLY_RIGHT_BY_LEFT_BUTTON_INDEX];
					
				} else if (Button_hitTest(&buttons[SWAP_BUTTON_INDEX], x, y)) {
					selectedButton = &buttons[SWAP_BUTTON_INDEX];
					
				} else if (Button_hitTest(&buttons[RESET_OPERAND_TO_IDENTITY_BUTTON_INDEX], x, y)) {
					selectedButton = &buttons[RESET_OPERAND_TO_IDENTITY_BUTTON_INDEX];
					
				} else if (Button_hitTest(&buttons[SHOW_FUNCTIONS_BUTTON_INDEX], x, y)) {
					selectedButton = &buttons[SHOW_FUNCTIONS_BUTTON_INDEX];
					
				} else if (Button_hitTest(&buttons[ROTATE_MULTIPLY_ORDER_BUTTON_INDEX], x, y)) {
					selectedButton = &buttons[ROTATE_MULTIPLY_ORDER_BUTTON_INDEX];
				}
			}
		} else {
			if (draggingNumberField) {
				MatrixControls_mouseUp(x, y);
				draggingNumberField = false;
				
			} else if (draggingFunctionField) {
				FunctionControls_mouseUp(x, y);
				draggingFunctionField = false;
				
			} else if (selectedButton != NULL) {
				if (Button_hitTest(selectedButton, x, y)) {
					if (selectedButton == &buttons[RESET_TO_IDENTITY_BUTTON_INDEX]) {
						matrix = Matrix_identity();
						MatrixControls_setMainMatrix(matrix);
						
					} else if (selectedButton == &buttons[SHOW_FUNCTIONS_BUTTON_INDEX]) {
						showingFunctions = !showingFunctions;
						buttons[SHOW_FUNCTIONS_BUTTON_INDEX].label = (showingFunctions ? " Numbers ->" : "Functions ->");
						
					} else if (selectedButton == &buttons[MULTIPLY_MODE_BUTTON_INDEX]) {
						autoMultiplyMode++;
						autoMultiplyMode %= 3;
						buttons[MULTIPLY_MODE_BUTTON_INDEX].label = (autoMultiplyMode == 0 ? " Show:  left only" : autoMultiplyMode == 1 ? "Show: left * right" : "Show: right * left");
						
					} else if (selectedButton == &buttons[MULTIPLY_LEFT_BY_RIGHT_BUTTON_INDEX]) {
						Matrix_multiply(&matrix, MatrixControls_getOperandMatrix());
						MatrixControls_setMainMatrix(matrix);
						
					} else if (selectedButton == &buttons[MULTIPLY_RIGHT_BY_LEFT_BUTTON_INDEX]) {
						matrix = Matrix_multiplied(MatrixControls_getOperandMatrix(), matrix);
						MatrixControls_setMainMatrix(matrix);
						
					} else if (selectedButton == &buttons[SWAP_BUTTON_INDEX]) {
						Matrix operandMatrix;
						
						operandMatrix = MatrixControls_getOperandMatrix();
						MatrixControls_setOperandMatrix(matrix);
						matrix = operandMatrix;
						MatrixControls_setMainMatrix(matrix);
						
					} else if (selectedButton == &buttons[RESET_OPERAND_TO_IDENTITY_BUTTON_INDEX]) {
						MatrixControls_setOperandMatrix(Matrix_identity());
						
					} else if (selectedButton == &buttons[ROTATE_MULTIPLY_ORDER_BUTTON_INDEX]) {
						multiplyRotationFirst = !multiplyRotationFirst;
						buttons[ROTATE_MULTIPLY_ORDER_BUTTON_INDEX].label = (multiplyRotationFirst ? "Drag mode: rotation * current" : "Drag mode: current * rotation");
						
					} else if (selectedButton == &buttons[APPLY_TRANSLATE_BUTTON_INDEX]) {
						Matrix_multiply(&matrix, FunctionControls_getTranslationMatrix());
						MatrixControls_setMainMatrix(matrix);
						
					} else if (selectedButton == &buttons[APPLY_SCALE_BUTTON_INDEX]) {
						Matrix_multiply(&matrix, FunctionControls_getScalingMatrix());
						MatrixControls_setMainMatrix(matrix);
						
					} else if (selectedButton == &buttons[APPLY_ROTATE_BUTTON_INDEX]) {
						Matrix_multiply(&matrix, FunctionControls_getRotationMatrix());
						MatrixControls_setMainMatrix(matrix);
						
					} else if (selectedButton == &buttons[APPLY_SHEARX_BUTTON_INDEX]) {
						Matrix_multiply(&matrix, FunctionControls_getShearXMatrix());
						MatrixControls_setMainMatrix(matrix);
						
					} else if (selectedButton == &buttons[APPLY_SHEARY_BUTTON_INDEX]) {
						Matrix_multiply(&matrix, FunctionControls_getShearYMatrix());
						MatrixControls_setMainMatrix(matrix);
						
					} else if (selectedButton == &buttons[APPLY_SHEARZ_BUTTON_INDEX]) {
						Matrix_multiply(&matrix, FunctionControls_getShearZMatrix());
						MatrixControls_setMainMatrix(matrix);
						
					} else if (selectedButton == &buttons[APPLY_PERSPECTIVE_BUTTON_INDEX]) {
						Matrix_multiply(&matrix, FunctionControls_getPerspectiveMatrix());
						MatrixControls_setMainMatrix(matrix);
						
					} else if (selectedButton == &buttons[APPLY_INVERT_BUTTON_INDEX]) {
						Matrix_invert(&matrix);
						MatrixControls_setMainMatrix(matrix);
						
					} else if (selectedButton == &buttons[APPLY_TRANSPOSE_BUTTON_INDEX]) {
						Matrix_transpose(&matrix);
						MatrixControls_setMainMatrix(matrix);
					}
				}
				selectedButton = NULL;
			}
		}
		glutPostRedisplay();
	}
}

#define ROTATION_RADIANS_PER_PIXEL (M_PI / 400.0)

void motionFunc(int x, int y) {
	if (mouseIsDown) {
		GLint viewport[4];
		int deltaX, deltaY;
		
		glGetIntegerv(GL_VIEWPORT, viewport);
		y = viewport[3] - y;
		
		deltaX = x - lastMouseX;
		deltaY = y - lastMouseY;
		lastMouseX = x;
		lastMouseY = y;
		
		if (draggingNumberField) {
			MatrixControls_mouseDragged(deltaX, deltaY);
			
		} else if (draggingFunctionField) {
			FunctionControls_mouseDragged(deltaX, deltaY);
			
		} else if (selectedButton == NULL) {
			Vector up = {0.0f, 1.0f, 0.0f}, right = {-1.0f, 0.0f, 0.0f};
			
			if (multiplyRotationFirst) {
				matrix = Matrix_multiplied(Matrix_rotated(Matrix_identity(), up,    deltaX * ROTATION_RADIANS_PER_PIXEL), matrix);
				matrix = Matrix_multiplied(Matrix_rotated(Matrix_identity(), right, deltaY * ROTATION_RADIANS_PER_PIXEL), matrix);
				
			} else {
				Matrix inverseRotation;
				
				inverseRotation = matrix;
				inverseRotation.m[3] = 0.0f;
				inverseRotation.m[7] = 0.0f;
				inverseRotation.m[11] = 0.0f;
				inverseRotation.m[12] = 0.0f;
				inverseRotation.m[13] = 0.0f;
				inverseRotation.m[14] = 0.0f;
				inverseRotation.m[15] = 1.0f;
				Matrix_transpose(&inverseRotation);
				up = Matrix_multiplyVector(inverseRotation, up);
				if (Vector_magnitudeSquared(up) > 0.0001f) {
					Matrix_rotate(&matrix, Vector_normalized(up), deltaX * ROTATION_RADIANS_PER_PIXEL);
				}
				
				inverseRotation = matrix;
				inverseRotation.m[3] = 0.0f;
				inverseRotation.m[7] = 0.0f;
				inverseRotation.m[11] = 0.0f;
				inverseRotation.m[12] = 0.0f;
				inverseRotation.m[13] = 0.0f;
				inverseRotation.m[14] = 0.0f;
				inverseRotation.m[15] = 1.0f;
				Matrix_transpose(&inverseRotation);
				right = Matrix_multiplyVector(inverseRotation, right);
				if (Vector_magnitudeSquared(right) > 0.0001f) {
					Matrix_rotate(&matrix, Vector_normalized(right), deltaY * ROTATION_RADIANS_PER_PIXEL);
				}
			}
			
			MatrixControls_setMainMatrix(matrix);
			glutPostRedisplay();
		}
	}
}

void keyboardFunc(unsigned char key, int x, int y) {
	if (key == 'r') {
		Matrix_loadIdentity(&matrix);
		MatrixControls_setMainMatrix(matrix);
		glutPostRedisplay();
	} else if (key == ' ') {
		if (draggingNumberField) {
			MatrixControls_setEditFieldValue(0.0f);
		} else if (draggingFunctionField) {
			FunctionControls_setEditFieldValue(0.0f);
		}
	} else if (key >= '0' && key <= '9') {
		if (draggingNumberField) {
			MatrixControls_setEditFieldValue(key - '0');
		} else if (draggingFunctionField) {
			FunctionControls_setEditFieldValue(key - '0');
		}
	} else if (key == '-') {
		if (draggingNumberField) {
			MatrixControls_setEditFieldValue(-MatrixControls_getEditFieldValue());
		} else if (draggingFunctionField) {
			FunctionControls_setEditFieldValue(-FunctionControls_getEditFieldValue());
		}
	}
}

int main(int argc, char ** argv) {
	glutInit(&argc, argv);
	
	glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE | GLUT_DEPTH);
	glutInitWindowPosition(2, 28);
	glutInitWindowSize(1000, 600);
	glutCreateWindow("Matrix playground");
	
	glutReshapeFunc(reshapeFunc);
	glutDisplayFunc(displayFunc);
	
	glutMouseFunc(mouseFunc);
	glutMotionFunc(motionFunc);
	glutKeyboardFunc(keyboardFunc);
	
	initGL();
	initPlayground();
	
	glutMainLoop();
	
	return EXIT_SUCCESS;
}
